// ---------------------------------------------------------------------------------------------------------------------------- \\
//															        \\
// File: Life.fth       Author: Oliver SINGLA	 Date: Dec, 27, 94  Version: 1.00					        \\
// ----		        ------			 ----		    -------						        \\
//															        \\
// Description:	 CONWAY's life game.											        \\
//															        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\

( Force decimal for integer numeric I/O )
decimal

( Compiler strategy )
?speed true!						       // Choose speed vs. size
?safe true!						       // Run-time check stack under/overflow
?iarithoverflow true!					       // Run-time check integer arithmetic overflow
?chat true!						       // Print a dot for each compiled definition

forget <task>
: <task> ;

( Make SCREEN the context vocabulary )
screen

( Colors used )
black LightGray BackForeGround const C_NormalText	       // Normal text
DarkGray const CellOffColor				       // Color of no cell (graphics mode)
   white const CellOnColor				       // Color of cell	   (graphics mode)

( Constants used )
640 const XSIZE						       // Max width
464 const YSIZE						       // Max height
 16 const XINC						       // X increment inside cells editor
  8 const YINC						       // Y increment inside cells editor
   0 const EVEN						       // Even generation (2, 4, ... )
   1 const ODD						       // Odd  generation (1, 3, ... )
'. const char_cell_off					       // Character used as cell off inside editor
'o const char_cell_on					       // Character used as cell on  inside editor
0x07 const char_cell_off_curr				       // Character used as cell off inside editor
0x0a const char_cell_on_curr				       // Character used as cell on  inside editor

( Keys used )
FKeys
KEY_F1  const KEY_HELP					       // F1  : Help
KEY_F2  const KEY_NEXT_GEN				       // F2  : Compute and show next generation
KEY_F3  const KEY_AUTO					       // F3  : Automatic mode (flip flop)
KEY_F10 const KEY_EDIT					       // F4  : Allows to edit the cells
KEY_ESC const KEY_BYE					       // ESC : Bye application
0x20    const KEY_CHGCELL				       // Change the state of the cell under the cursor (CellsEditor)
Screen

( Global variables )
int gnGenerationNo					       // Generation (1, 2, ... )
int gnPopulationsNb					       // Nb of living cells
int gnEditor_X						       // Current column within editor
int gnEditor_Y						       // Current row within editor
int gnCornerTopLeft_X					       // Column of the top/left corner
int gnCornerTopLeft_Y					       // Row of the top/left corner
logical glEditor?					       // Is editor active ?


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : InfosLineNo												        \\
// 	 Stack : ( --- n )												        \\
// Description : Return the 1st row of informations.									        \\
//    Callers : ShowGenPop ShowAvailableKeysCmd Frame ShowGeneration NextGeneration:2 AutomaticMode Help:2 CellsEditor:3       \\
// 	 Calls :													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  InfosLineNo
   #rows 1-
   ;   // InfosLineNo


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : CellsWorld												        \\
// 	 Stack : ( --- )												        \\
// Description : Creating word for Cells										        \\
//    Callers : cells													        \\
// 	 Calls : XSIZE YSIZE												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  CellsWorld
   build>
       XSIZE YSIZE *  ballot				       // Allocate space for each cells
       drop						       // ( addr --- )
   does>						       // ( x y addr1 --- addr2 )
//     swap XSIZE * + +
       ( mov   eax, [esi + 4] )	 [: 8B 46 04 :]
       ( imul  eax, XSIZE     )	 [: 69 C0 80 02 00 00 :]
       ( add   eax, [esi]     )	 [: 03 06 :]
       ( add   esi, 8	      )	 [: 83 C6 08 :]
       ( add   [esi], eax     )	 [: 01 06 :]
   ;   // CellsWorld


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Cells													        \\
// 	 Stack : RunTime: ( x y -- addr )										        \\
// Description : Return address for the cell (x, y)    (x is the width, y is the height)				        \\
//    Callers : PutCell GetCell RvsCell InitThings ZeroLife:5 ComputeNextGeneration ShowGeneration Editor_ShowGeneration       \\
// 		 Editor_ShowCells											        \\
// 	 Calls : CellsWorld												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
dynamic CellsWorld cells


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : PutCell												        \\
// 	 Stack : ( n x y --- )												        \\
// Description : Store a living cell (1) or dead cell (0), at column <x>, row <y>.					        \\
//    Callers :													        \\
// 	 Calls : cells													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  PutCell
   Cells						       // ( n x y --- n addr )
   swap							       // ( n addr --- addr n )
   gnGenerationNo @ 2 mod				       // ( addr n --- addr n 0|1 )
   swap							       // ( addr n 0|1 --- addr 0|1 n )
   if
       swap SetBit					       // ( n addr --- )
   else
       swap ResetBit					       // ( n addr --- )
   endif
   ;   // PutCell


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : GetCell												        \\
// 	 Stack : ( x y --- n )												        \\
// Description : Return 0 (dead cell) or 1 (living cell), from column <x>, row <y>.					        \\
//    Callers : changes> Editor_ChangeCell:2 CellsEditor:2								        \\
// 	 Calls : cells													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  GetCell
   Cells						       // ( x y --- addr )
   gnGenerationNo @ 2 mod				       // ( addr --- addr 0|1 )
   swap							       // ( addr 0|1 --- 0|1 addr )
   GetBit						       // ( n1 addr --- n2 )
   ;   // GetCell


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : RvsCell												        \\
// 	 Stack : ( x y --- )												        \\
// Description : Reverse the Dead/Living status of the cell, at column <x>, row <y>.					        \\
//    Callers : Editor_ChangeCell											        \\
// 	 Calls : cells													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  RvsCell
   Cells						       // ( x y --- addr )
   gnGenerationNo @ 2 mod				       // ( addr --- addr 0|1 )
   swap							       // ( addr 0|1 --- 0|1 addr )
   ComplBit						       // ( n addr --- )
   ;   // RvsCell


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : changes>												        \\
// 	 Stack : CompileTime: ( --- )											        \\
// 		     RunTime: ( n --- x y f )										        \\
// Description : Create the array of which changed after a computation.							        \\
//    Callers : changes												        \\
// 	 Calls : XSIZE:2 YSIZE GetCell											        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  changes>
   build>
       XSIZE YSIZE * WordSize*				       // Room for: XY (dword)
       ballot drop					       // Ask for room
   does>
       over -1 == if
	   under
       else
	   swap WordSize*  +				       // ( n addr1 --- addr2 )
	   @ XSIZE /mod					       // ( addr --- x y )
	   ddup getcell					       // ( x y --- x y f )
       endif
   ;   // changes>


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : changes												        \\
// 	 Stack : ( n --- x y f )											        \\
// Description : Ask the array of which changed (updated by ComputeGeneration)						        \\
//    Callers : InitThings ComputeNextGeneration ShowGeneration life							        \\
// 	 Calls : changes>												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
dynamic changes> changes


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : InitThings												        \\
// 	 Stack : ( --- )												        \\
// Description : Initialize the game.											        \\
//    Callers : life													        \\
// 	 Calls : XSIZE YSIZE cells changes										        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  InitThings

   ( Allocate the huge array to store each cell state )
   alloc cells
   alloc changes

   ( Set initial editing position, the center of the array )
   XSIZE #cols - 2/  gnCornerTopLeft_X !
   YSIZE #rows - 2/  gnCornerTopLeft_Y !
   #cols 2/  gnCornerTopLeft_X @  +  gnEditor_X !
   #rows 2/  gnCornerTopLeft_Y @  +  gnEditor_Y !

   ( Initial population is empty )
   gnPopulationsNb 0!

   ( 1st generation )
   gnGenerationNo 1!

   ( Is editor active ? )
   glEditor? true!

   ;   // InitThings


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ShowGenPop												        \\
// 	 Stack : ( --- )												        \\
// Description : Print the generation # and the nb of populations.							        \\
//    Callers : Frame NextGeneration Editor_ChangeCell CellsEditor							        \\
// 	 Calls : InfosLineNo												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ShowGenPop
   InfosLineNo
    4 over gotoXY   6  gnPopulationsNb @  .R
   14 swap gotoXY   4  gnGenerationNo @	  .R
   ;   // ShowGenPop


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ShowAvailableKeysCmd											        \\
// 	 Stack : ( --- )												        \\
// Description : Draw the available keys.										        \\
//    Callers : Frame AutomaticMode CellsEditor									        \\
// 	 Calls : InfosLineNo												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ShowAvailableKeysCmd
   '~ >xylstring_char
   0 black white       >xylstring_attr
   1 black lightgray   >xylstring_attr
   InfosLineNo 19 swap					       // ( --- x y )
   glEditor? true? if
	   "  ~{0}F1~{1} help ~{0}SPACE~{1} on/off (shift)~{0}\18\19\1b\1a~{1} mov ~{0}F10~{1} ExitEditor ~{0}ESC~{1} bye"
   else
	   "  ~{0}F1~{1} help ~{0}F2~{1} next generation ~{0}F3~{1} automatic ~{0}F10~{1} editor ~{0}ESC~{1} bye"
   endif
   !xylstring
   clreol						       // Clear until end of line
   ;  // ShowAvailableKeysCmd


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Frame													        \\
// 	 Stack : ( --- )												        \\
// Description : Print the frame of the game.										        \\
//    Callers : Help CellsEditor life											        \\
// 	 Calls : InfosLineNo ShowGenPop ShowAvailableKeysCmd								        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Frame
   InfosLineNo
    1 over " Pop " !xystring
   11 swap " Gen " !xystring
   ShowAvailableKeysCmd					       // Draw the availble keys.
   ShowGenPop						       // Print the generation # and the nb of populations.
   ;  // Frame


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ZeroLife												        \\
// 	 Stack : ( --- )												        \\
// Description : Empty all the cells.											        \\
//    Callers : life													        \\
// 	 Calls : XSIZE:2 YSIZE:2 cells:5										        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ZeroLife

  //  0	 XSIZE 1-  make					       // For each column
  //	  0  YSIZE 1-  make				       // For each row
  //	     I J Cells  0C!
  //	  endmake
  //  endmake

   ( Reset all the field )
   0 0 Cells
   ( push	  edi		   )	     [: 57 :]
   ( clr	  al		   )	     [: 32 C0 :]
   ( mov	  edi, [esi]	   )	     [: 8B 3E :]
   ( add	  esi, 4	   )	     [: 83 C6 04 :]
   ( mov	  ecx, XSIZE*YSIZE )	     [: B9 48800 :]
   ( cld			   )	     [: FC :]
   ( rep	  stosb		   )	     [: F3 AA :]
   ( pop	  edi		   )	     [: 5F :]

   ( Mark border cells as inactive {bit 7 is set} )
   0  XSIZE 1-  make
       7  I  0	       Cells  SetBit			       // First line
       7  I  YSIZE 1-  Cells  SetBit			       // Last line
   endmake
   0  YSIZE 1-  make
       7  0	    I  Cells  SetBit			       // Fist column
       7  XSIZE 1-  I  Cells  SetBit			       // Last column
   endmake

   ( Nobody is alive ! )
   gnPopulationsNb 0!

   ;  // ZeroLife


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ComputeEvenGeneration											        \\
// 	 Stack : ( --- )												        \\
// Description : Compute generation from EVEN generation to ODD generation.						        \\
//    Callers : ComputeNextGeneration											        \\
// 	 Calls :													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
: ComputeEvenGeneration

   (	        push   esi			 )      [: 56 :]
   (	        mov    esi, [esi]		 )      [: 8B 36 :]
   (	        clr    ecx			 )      [: 33 C9 :]
   (	        mov    edx, 640 * 464 - 640	 )      [: BA 80 85 04 00 :]
   (	        dec    edx			 )      [: 4A :]

   (   loop:    test   byte ptr [esi], 80h	 )      [: F6 06 80 :]
   (	        jnz    @@NextCell		 )      [: 75 63 :]

   ( Compute in BL the number of neighbors for this cell )
   (		 clr    ebx			 )      [: 33 DB :]
   (		 mov    eax, [esi - 641]	 )      [: 8B 86 7F FD FF FF :]
   (		 bt     eax, 16			 )      [: 0F BA E0 10 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 bt     eax, 8			 )      [: 0F BA E0 08 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 bt     eax, 0			 )      [: 0F BA E0 00 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 mov    eax, [esi - 1]		 )      [: 8B 46 FF :]
   (		 bt     eax, 16			 )      [: 0F BA E0 10 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 bt     eax, 0			 )      [: 0F BA E0 00 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 mov    eax, [esi + 639]	 )      [: 8B 86 7F 02 00 00 :]
   (		 bt     eax, 0			 )      [: 0F BA E0 00 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 bt     eax, 8			 )      [: 0F BA E0 08 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]
   (		 bt     eax, 16			 )      [: 0F BA E0 10 :]
   (		 adc    bl, 0			 )      [: 80 D3 00 :]

   (		 cmp    bl, 2			 )      [: 80 FB 02 :]
   (		 jl     dead			 )      [: 7C 12 :]
   (		 cmp    bl, 3			 )      [: 80 FB 03 :]
   (		 jg     dead			 )      [: 7F 0D :]
   (		 je     living			 )      [: 74 05 :]
   (		 test   byte ptr [esi], 01h	 )      [: F6 06 01 :]
   (		 jz     dead			 )      [: 74 06 :]

   (   living:	 or     byte ptr [esi], 02h	 )      [: 80 0E 02 :]
   (		 inc    ecx			 )      [: 41 :]
   (		 jmp    NextCell		 )      [: EB 03 :]

   (   dead:	 and    byte ptr [esi], 0FDh	 )      [: 80 26 FD :]

   (   NextCell: inc    esi			 )      [: 46 :]
   (		 dec    edx			 )      [: 4A :]
   (		 jne    loop			 )      [: 75 94 :]
   (		 pop    esi			 )      [: 5E :]

   ( Clear previous generation bit )
   (	        push   edi		        )      [: 57 :]
   (	        push   ecx		        )      [: 51 :]
   (	        push   esi		        )      [: 56 :]
   (	        mov    ebx, [esi+4]	        )      [: 8B 5E 04 :]
   (	        mov    esi, [esi]	        )      [: 8B 36 :]
   (	        clr    ecx		        )      [: 33 C9 :]
   (	        clr    edx		        )      [: 33 D2 :]
   (	        clr    edi		        )      [: 33 FF :]
   (   ClrPrevG: mov    al, [esi]	        )      [: 8A 06 :]
   (	        and    al, 3		        )      [: 24 03 :]
   (	        jz     NoChang		        )      [: 74 10 :]
   (	        cmp    al, 3		        )      [: 3C 03 :]
   (	        je     NoChang		        )      [: 74 0C :]
   (   Change:  and    byte ptr [esi], 82h      )      [: 80 26 82 :]
   (	        or     byte ptr [esi], 04h      )      [: 80 0E 04 :]
   (	        mov    [ebx + edx*4], edx       )      [: 89 14 BB :]
   (	        inc    edi		        )      [: 47 :]
   (	        jmp    Done		        )      [: EB 03 :]
   (   NoChang: and    byte ptr [esi], 82h      )      [: 80 26 82 :]
   (   Done:    inc    esi		        )      [: 46 :]
   (	        inc    edx		        )      [: 42 :]
   (	        cmp    edx, 640*464	        )      [: 81 FA 00 88 04 00 :]
   (	        jne    ClrPrevG		        )      [: 75 DD :]
   (	        pop    esi		        )      [: 5E :]
   (	        mov    [esi+4], edi	        )      [: 89 7E 04 :]
   (	        pop    ecx		        )      [: 59 :]
   (	        mov    [esi], ecx	        )      [: 89 0E :]
   (	        pop    edi		        )      [: 5F :]

   ;  // ComputeEvenGeneration


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ComputeOddGeneration											        \\
// 	 Stack : ( addr1 addr2 --- n1 n2 )										        \\
// 		     <addr1> : base address of changes									        \\
// 		     <addr2> : base address of cells									        \\
// 		     <n1>    : number of cells which have changed (stored in changes array).				        \\
// 		     <n2>    : new number of cells									        \\
// Description : Compute generation from ODD generation to EVEN generation.						        \\
//    Callers : ComputeNextGeneration											        \\
// 	 Calls :													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
: ComputeOddGeneration

   (	        push   esi		        )      [: 56 :]
   (	        mov    esi, [esi]	        )      [: 8B 36 :]
   (	        clr    ecx		        )      [: 33 C9 :]
   (	        mov    edx, 640 * 464 - 640     )      [: BA 80 85 04 00 :]
   (	        dec    edx		        )      [: 4A :]

   (   loop:    test   byte ptr [esi], 80h      )      [: F6 06 80 :]
   (	        jnz    @@NextCell	        )      [: 75 63 :]

   ( Compute in BL the number of neighbors for this cell )
   (	        clr    ebx		        )      [: 33 DB :]
   (	        mov    eax, [esi - 641]	        )      [: 8B 86 7F FD FF FF :]
   (	        bt     eax, 17		        )      [: 0F BA E0 11 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        bt     eax, 9		        )      [: 0F BA E0 09 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        bt     eax, 1		        )      [: 0F BA E0 01 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        mov    eax, [esi - 1]	        )      [: 8B 46 FF :]
   (	        bt     eax, 17		        )      [: 0F BA E0 11 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        bt     eax, 1		        )      [: 0F BA E0 01 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        mov    eax, [esi + 639]	        )      [: 8B 86 7F 02 00 00 :]
   (	        bt     eax, 1		        )      [: 0F BA E0 01 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        bt     eax, 9		        )      [: 0F BA E0 09 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]
   (	        bt     eax, 17		        )      [: 0F BA E0 11 :]
   (	        adc    bl, 0		        )      [: 80 D3 00 :]

   (	        cmp    bl, 2		        )      [: 80 FB 02 :]
   (	        jl     dead		        )      [: 7C 12 :]
   (	        cmp    bl, 3		        )      [: 80 FB 03 :]
   (	        jg     dead		        )      [: 7F 0D :]
   (	        je     living		        )      [: 74 05 :]
   (	        test   byte ptr [esi], 02h      )      [: F6 06 02 :]
   (	        jz     dead		        )      [: 74 06 :]

   (   living:  or     byte ptr [esi], 01h      )      [: 80 0E 01 :]
   (	        inc    ecx		        )      [: 41 :]
   (	        jmp    NextCell		        )      [: EB 03 :]

   (   dead:    and    byte ptr [esi], 0FEh     )      [: 80 26 FE :]

   (  NextCell: inc    esi		        )      [: 46 :]
   (	        dec    edx		        )      [: 4A :]
   (	        jne    loop		        )      [: 75 94 :]
   (	        pop    esi		        )      [: 5E :]

   ( Clear previous generation bit )
   (	        push   edi		        )      [: 57 :]
   (	        push   ecx		        )      [: 51 :]
   (	        push   esi		        )      [: 56 :]
   (	        mov    ebx, [esi+4]	        )      [: 8B 5E 04 :]
   (	        mov    esi, [esi]	        )      [: 8B 36 :]
   (	        clr    ecx		        )      [: 33 C9 :]
   (	        clr    edx		        )      [: 33 D2 :]
   (	        clr    edi		        )      [: 33 FF :]
   (  ClrPrevG: mov    al, [esi]	        )      [: 8A 06 :]
   (	        and    al, 3		        )      [: 24 03 :]
   (	        jz     NoChange		        )      [: 74 10 :]
   (	        cmp    al, 3		        )      [: 3C 03 :]
   (	        je     NoChange		        )      [: 74 0C :]
   (   Change:  and    byte ptr [esi], 81h      )      [: 80 26 81 :]
   (	        or     byte ptr [esi], 04h      )      [: 80 0E 04 :]
   (	        mov    [ebx + edx*4], edx       )      [: 89 14 BB :]
   (	        inc    edi		        )      [: 47 :]
   (	        jmp    Done		        )      [: EB 03 :]
   (  NoChange: and    byte ptr [esi], 81h      )      [: 80 26 81 :]
   (   Done:    inc    esi		        )      [: 46 :]
   (	        inc    edx		        )      [: 42 :]
   (	        cmp    edx, 640*464	        )      [: 81 FA 00 88 04 00 :]
   (	        jne    ClrPrevG		        )      [: 75 DD :]
   (	        pop    esi		        )      [: 5E :]
   (	        mov    [esi+4], edi	        )      [: 89 7E 04 :]
   (	        pop    ecx		        )      [: 59 :]
   (	        mov    [esi], ecx	        )      [: 89 0E :]
   (	        pop    edi		        )      [: 5F :]

   ;  // ComputeOddGeneration


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ComputeNextGeneration											        \\
// 	 Stack : ( --- n )												        \\
// Description : Compute next generation. Return the new nb of cells.							        \\
//    Callers : NextGeneration												        \\
// 	 Calls : cells changes ComputeEvenGeneration ComputeOddGeneration						        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ComputeNextGeneration
   -1 changes						       // ( --- addr1 )
   0 0 cells						       // ( addr1 --- addr1 addr2 )
   gnGenerationNo @ 2 mod if
       ComputeOddGeneration
   else
       ComputeEvenGeneration
   endif
   ;  // ComputeNextGeneration


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ShowGeneration												        \\
// 	 Stack : ( f --- )  <f>: force all cells to be redrawn								        \\
// Description : Show cells which have changed.										        \\
//    Callers : NextGeneration Help CellsEditor									        \\
// 	 Calls : CellOffColor CellOnColor:2 gnPopulationsNb InfosLineNo cells changes					        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
graphics
:  ShowGeneration
   not if

       ( Show only cells which have changed )
       dup not if
	   drop
       else
	   1- 0 swap make
	       I changes				       // ( --- x y f )
	       if CellOnColor else CellOffColor endif	       // ( x y f --- x y n )
	       setpixel					       // ( x y n --- )
	   endmake
       endif

   else

       ( Show all cells )
       int nCellsNb  gnPopulationsNb nCellsNb @!	       // Nb of cells to draw
       0  #gcols 1-  make
	   I 8 mod 0== if
	       screen
	       I 8 /  InfosLineNo  '.  !xychar
	       graphics
	   endif
	   0 #grows 17 -	 make
	       gnGenerationNo @ 2 mod
	       J I cells	 GetBit	 if
		   J I CellOnColor SetPixel
		   nCellsNb 1-!
	       endif
	   endmake
	   nCellsNb @ when
       endmake

   endif
   ;  // ShowGeneration
screen


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_ShowGeneration											        \\
// 	 Stack : ( --- )												        \\
// Description : Show cells which have changed.										        \\
//    Callers :													        \\
// 	 Calls : cells													        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_ShowGeneration
   ?UpdateCursor false!					       // Don't update cursor position when drawing
   0 #rows 3 -  make
       I 0 gotoxy
       0  #cols 1-  make
	   gnCornerTopLeft_X @ I +  gnCornerTopLeft_Y @ J +  Cells // ( --- addr )
	   2 over  getbit
	   if						       // ( addr b1 b2 --- addr )
	       gnGenerationNo @ 2 mod  swap  GetBit	       // ( addr --- b )
	       if					       // ( b --- )
		   J I " o" !xystring
	       else
		   J I " ." !xystring
	       endif
	   else
	       drop
	   endif
       endmake
   endmake
   ?UpdateCursor true!					       // Update cursor position when drawing
   ;  // Editor_ShowGeneration


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_ShowCells											        \\
// 	 Stack : ( --- )												        \\
// Description : Show generation, in editor mode (text mode).								        \\
//    Callers : Editor_Down Editor_Up Editor_Left Editor_Right Help CellsEditor					        \\
// 	 Calls : char_cell_off char_cell_on cells									        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_ShowCells
   ?UpdateCursor false!					       // Don't update cursor position when drawing
   0 #rows 2 -  make
       0 I gotoxy
       0  #cols 1-  make
	   gnGenerationNo @ 2 mod
	   gnCornerTopLeft_X @ I +  gnCornerTopLeft_Y @ J +  Cells
	   getbit if char_cell_on else char_cell_off endif
	   emit
       endmake
   endmake
   ?UpdateCursor true!					       // Update cursor position when drawing
   ;  // Editor_ShowCells


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : NextGeneration												        \\
// 	 Stack : ( --- )												        \\
// Description : Compute and show next generation of cells.								        \\
//    Callers : AutomaticMode HeartLife										        \\
// 	 Calls : InfosLineNo:2 ShowGenPop ComputeNextGeneration ShowGeneration						        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  NextGeneration
   0 InfosLineNo red 'C !xyachar
   ComputeNextGeneration				       // ( --- n1 n2) n1: nb of changed cells n2: new population
   0 InfosLineNo 0x20 !xychar
   gnPopulationsNb !					       // ( n --- ) Update the nb of cells
   gnGenerationNo 1+!					       // Next generation
   ShowGenPop						       // Redraw the nb of cells and the generation nb
   FALSE ShowGeneration					       // Show only cells which have changed
   ;  // NextGeneration


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : AutomaticMode												        \\
// 	 Stack : ( --- )												        \\
// Description : Enter into automatic mode.										        \\
//    Callers : HeartLife												        \\
// 	 Calls : KEY_AUTO InfosLineNo ShowAvailableKeysCmd NextGeneration						        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  AutomaticMode
   '~ >xylstring_char
   0 black white     >xylstring_attr
   1 black lightgray >xylstring_attr
   InfosLineNo 19 swap					       // ( --- x y )
   "   ~{0}F3~{1} cancel automatic mode "
   !xylstring clreol					       // ( x y addrz --- )
   logical lStop?  lStop? false!			       // Exit loop ?
   begin
       NextGeneration
       begin
	   lStop? false?
	   ?key
	   and
       while
	   key KEY_AUTO ==  if
	       lStop? true!
	   else
	       lStop? false!
	   endif
       repeat
       lStop? true?
   until
   ShowAvailableKeysCmd					       // Draw the available keys.
   ;   // AutomaticMode


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_ChangeCell											        \\
// 	 Stack : ( --- )												        \\
// Description : Change the Living/Death status of the current cell.							        \\
//    Callers : CellsEditor												        \\
// 	 Calls : char_cell_off char_cell_on GetCell:2 RvsCell ShowGenPop						        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_ChangeCell

   gnEditor_X @	 gnEditor_Y @  RvsCell			       // Change the state of the cell under the cursor
   gnEditor_X @	 gnCornerTopLeft_X @  -			       // ( --- col )
   gnEditor_Y @	 gnCornerTopLeft_Y @  -			       // ( col --- col row )
   gnEditor_X @	 gnEditor_Y @  GetCell			       // ( col row --- col row n )
   if  char_cell_on  else  char_cell_off  endif
   !xychar

   ( Update the nb of cells )
   gnEditor_X @	 gnEditor_Y @  GetCell  if
       gnPopulationsNb 1+!
   else
       gnPopulationsNb 1-!
   endif

   ( Draw the nb of cells )
   ShowGenPop

   ;  // Editor_ChangeCell


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_Down												        \\
// 	 Stack : ( --- )												        \\
// Description : Into cells editor, move cursor one line down.								        \\
//    Callers : CellsEditor												        \\
// 	 Calls : YSIZE:3 YINC Editor_ShowCells										        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_Down
   gnEditor_Y @	 YSIZE #rows 1- - !=  if
       gnEditor_Y @  gnCornerTopLeft_Y @  -  #rows 1- 1- !=  if
	   gnEditor_Y 1+!
       else
	   YINC gnCornerTopLeft_Y +!
	   gnEditor_Y @	 YSIZE #rows 1- - >  if
	       YSIZE #rows 1- -	 gnCornerTopLeft_Y !
	   endif
	   Editor_ShowCells				       // Show cells in text mode
       endif
   endif
   ;  // Editor_Down


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_Up												        \\
// 	 Stack : ( --- )												        \\
// Description : Into cells editor, move cursor one line up.								        \\
//    Callers : CellsEditor												        \\
// 	 Calls : YINC Editor_ShowCells											        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_Up
   gnEditor_Y @0!= if
       gnEditor_Y @  gnCornerTopLeft_Y @  != if
	   gnEditor_Y 1-!
       else
	   gnCornerTopLeft_Y @0!=  if
	       YINC gnCornerTopLeft_Y -!
	       gnCornerTopLeft_Y @ 0< if
		   gnCornerTopLeft_Y 0!
	       endif
	       Editor_ShowCells				       // Show cells in text mode
	   endif
       endif
   endif
   ;  // Editor_Up


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_Left												        \\
// 	 Stack : ( --- )												        \\
// Description : Into cells editor, move cursor one char left.								        \\
//    Callers : CellsEditor												        \\
// 	 Calls : XINC Editor_ShowCells											        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_Left
   gnEditor_X @0!= if
       gnEditor_X @  gnCornerTopLeft_X @  !=  if
	   gnEditor_X 1-!
       else
	   gnCornerTopLeft_X @0!=  if
	       XINC gnCornerTopLeft_X -!
	       gnCornerTopLeft_X @ 0< if
		   gnCornerTopLeft_X 0!
	       endif
	       Editor_ShowCells				       // Show cells in text mode
	   endif
       endif
   endif
   ;  // Editor_Left


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Editor_Right												        \\
// 	 Stack : ( --- )												        \\
// Description : Into cells editor, move one char right.								        \\
//    Callers : CellsEditor												        \\
// 	 Calls : XSIZE:3 XINC Editor_ShowCells										        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Editor_Right
   gnEditor_X @	 XSIZE #cols - !=  if
       gnEditor_X @  gnCornerTopLeft_X @  -  #cols 1- !=  if
	   gnEditor_X 1+!
       else
	   XINC gnCornerTopLeft_X +!
	   gnEditor_X @	 XSIZE #cols - >  if
	       XSIZE #cols -  gnCornerTopLeft_X !
	   endif
	   Editor_ShowCells				       // Show cells in text mode
       endif
   endif
   ;  // Editor_Right


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ShowHelp_1												        \\
// 	 Stack : ( --- )												        \\
// Description : Show life help, page 1											        \\
//    Callers : ShowHelp												        \\
// 	 Calls : CellOffColor:6												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ShowHelp_1
   CellOffColor graphics clear screen
   '~ >xylstring_char
   0 CellOffColor white	    >xylstring_attr
   1 CellOffColor lightgray >xylstring_attr
   2 CellOffColor yellow    >xylstring_attr
   34  0 " ~{1}LIFE GAME"                                                                                      !xylstring
   34  1 " ---------"                                                                                          !xylstring
    0  3 " ~{0}In the game of Life, cells are living on a large checkerboard and die or"                       !xylstring
    0  4 " ~{0}reproduce  according to some simple genetic rules. First note that each cell"                   !xylstring
    0  5 " ~{0}of the checkerboard (640x464) has ~{1}eight~{0} neighboring cells, four adjacent"               !xylstring
    0  6 " ~{0}orthogonally, four adjacent diagonally. The rules are:"                                         !xylstring
    0  7 "   ~{2}Survivals~{0}: Every cell with ~{1}two~{0} or ~{1}three~{0} neighbors survives for the next"  !xylstring
    0  8 " ~{0}             generation."                                                                       !xylstring
    0  9 "      ~{2}Deaths~{0}: Each cell with ~{1}four~{0} or more neighbors dies from"                       !xylstring
    0 10 "              ~{1}overpopulation~{0}. Every cell with ~{1}one~{0} neighbor or ~{1}none~{0} dies"     !xylstring
    0 11 " ~{0}             from ~{1}isolation~{0}."                                                           !xylstring
    0 12 "      ~{2}Births~{0}: Each empty cell adjacent to exactly ~{1}three~{0} neighbors"                   !xylstring
    0 13 " ~{0}             (no more, no fewer) is a ~{1}birth~{0} cell."                                      !xylstring
    0 14 " ~{0}You must understand that all births and deaths occur ~{1}simultaneous~{0}. Together"            !xylstring
    0 15 " ~{0}they constitute a single generation."                                                           !xylstring
    0 17 " You might try this very simple generation:   .........."                                            !xylstring
    0 18 "                                              ..xxx....."                                            !xylstring
    0 19 "                                              ..x..x...."                                            !xylstring
    0 20 "                                              ..x...x..."                                            !xylstring
    0 21 "                                              .........."                                            !xylstring

   '~ >xylstring_char
   0 CellOffColor white	    >xylstring_attr
   1 CellOffColor lightgray >xylstring_attr
   #cols 15 -  #rows 1- " ~{0}Press ~{1}ESC~{0} or ~{1}\19" !xylstring

   ;  // ShowHelp_1


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ShowHelp_2												        \\
// 	 Stack : ( --- )												        \\
// Description : Help, page 2												        \\
//    Callers : ShowHelp												        \\
// 	 Calls : CellOffColor:8												        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ShowHelp_2
   CellOffColor graphics clear screen
   0 const x0
   1 const y0
   '~ >xylstring_char
   0 CellOffColor lightgray >xylstring_attr
   1 CellOffColor red	    >xylstring_attr
   2 CellOffColor brown	    >xylstring_attr
   3 CellOffColor cyan	    >xylstring_attr
   4 CellOffColor white	    >xylstring_attr
   x0 y0  0 + " ~{0}This game is written in ~{1}FROTH~{0}.  This is a free implementation of"            !xylstring
   x0 y0  1 + " ~{1}FORTH~{0} language. Here are the highlights of FROTH language:"                      !xylstring
   x0 y0  3 + " ~{2}*~{4} Complete and highly interactive environment (compiler, editor, interpreter);"  !xylstring
   x0 y0  4 + " ~{2}*~{4} Run in 32-bits protected mode (flat model);"                                   !xylstring
   x0 y0  5 + " ~{2}*~{4} Tools for cross-reference (update each definition header);"                    !xylstring
   x0 y0  6 + " ~{2}*~{4} Definitions browser (instant go to any definition or variable);"               !xylstring
   x0 y0  7 + " ~{2}*~{4} Complete online help (help available for each definition);"                    !xylstring
   x0 y0  8 + " ~{2}*~{4} Up to 32 running threads (cooperative multi-tasking);"                         !xylstring
   x0 y0  9 + " ~{2}*~{4} Dynamic and automatic variables (with garbage collector);"                     !xylstring
   x0 y0 10 + " ~{2}*~{4} Fixed and Dynamic string handling;"                                            !xylstring
   x0 y0 11 + " ~{2}*~{4} Objects, classes and methods;"                                                 !xylstring
   x0 y0 12 + " ~{2}*~{4} Ability to build .EXE files (stand alone application under MS-DOS);"           !xylstring
   x0 y0 13 + " ~{2}*~{4} Comes with a complete 100 pages manual: tutorial and reference."               !xylstring
   x0 y0 15 + " ~{0}This is a ~{1}public domain~{0} software."                                           !xylstring
   x0 8 + #rows 3 - " ~{3}--- Oliver SINGLA - FROTH v 0.40.a - singla@email.teaser.fr ---"               !xylstring

   '~ >xylstring_char
   0 CellOffColor white	    >xylstring_attr
   1 CellOffColor lightgray >xylstring_attr
   #cols 15 -  #rows 1- " ~{0}Press ~{1}ESC~{0} or ~{1}\18" !xylstring

   ;  // ShowHelp_2


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : ShowHelp												        \\
// 	 Stack : ( --- )												        \\
// Description : Show life help.											        \\
//    Callers : Help													        \\
// 	 Calls : ShowHelp_1 ShowHelp_2											        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  ShowHelp
   int	   nPageNo     nPageNo 1!
   logical lStop?      lStop? false!
   begin
       nPageNo @ case
	   1 of	 ShowHelp_1  endof
	   2 of	 ShowHelp_2  endof
       endcase
       logical lOk?  lOk? false!
       begin
	   key case fkeys
	       0x1B	 of  lOk? true!	 lStop? true!			      endof
	       KEY_UP	 of  lOk? true!	 nPageNo @ 1 > if  nPageNo 1-! endif  endof
	       KEY_DOWN	 of  lOk? true!	 nPageNo @ 2 < if  nPageNo 1+! endif  endof
	   endcase
	   lOk? true?
       until
       lStop? true?
   until
   ;  // ShowHelp


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : Help													        \\
// 	 Stack : ( --- )												        \\
// Description : Show life help.											        \\
//    Callers : CellsEditor HeartLife											        \\
// 	 Calls : CellOffColor:2 InfosLineNo:2 Frame ShowGeneration Editor_ShowCells ShowHelp				        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  Help
   graphics CellOffColor clear screen			       // Erase graphics screen (set with off cells)
   ShowHelp						       // Show life help
   graphics CellOffColor clear screen			       // Erase graphics screen (set with off cells)
   InfosLineNo clrline					       // Put right color for this line
   glEditor? true? if
       Editor_ShowCells					       // Show cells in text mode
   else
       TRUE ShowGeneration				       // Show all cells, in graphics mode
       InfosLineNo clrline				       // Clear this line
   endif
   Frame						       // Redraw the 'frame' of the game
   ;  // Help


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : CellsEditor												        \\
// 	 Stack : ( --- )												        \\
// Description : Enter into the editor of cells.									        \\
//    Callers : HeartLife:2												        \\
// 	 Calls : CellOffColor char_cell_off char_cell_on char_cell_off_curr char_cell_on_curr KEY_HELP KEY_CHGCELL	        \\
// 		 InfosLineNo:3 GetCell:2 ShowGenPop ShowAvailableKeysCmd Frame ShowGeneration Editor_ShowCells		        \\
// 		 Editor_ChangeCell Editor_Down Editor_Up Editor_Left Editor_Right Help					        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  CellsEditor
   glEditor? true!					       // Is Editor active ?
   Editor_ShowCells					       // Show cells in text mode
   ShowAvailableKeysCmd					       // Draw the availble keys.
   logical lStop?   lStop? false!			       // Flag: stop main loop ?
   begin
       #cols 10 - InfosLineNo gotoXY
       '( emit  3 gnEditor_X @ .R  ', emit
       3 gnEditor_Y @ .R  ') emit			       // Print current (col, row)
       gnEditor_X @  gnCornerTopLeft_X @  -		       // ( --- x )	 Compute the current col
       gnEditor_Y @  gnCornerTopLeft_Y @  -		       // ( x --- x y )	 Compute the current row
       ddup gnEditor_X @ gnEditor_Y @ getcell		       // ( x y --- x y x y f )
       if char_cell_on_curr else char_cell_off_curr endif      // ( x y x y f --- x y x y n )
       !xychar						       // ( x y x y n --- x y )
       ddup gotoXY					       // ( x y --- x y )  Bring cursor on the current cell's position
       key						       // ( x y --- x y c )
       -rot						       // ( x y c --- c x y )
       gnEditor_X @ gnEditor_Y @ getcell		       // ( x y --- x y f )
       if char_cell_on else char_cell_off endif		       // ( x y f --- x y n )
       !xychar						       // ( x y x y n --- x y )
       FKeys case					       // ( c --- )
	   KEY_CHGCELL		 of  Editor_ChangeCell endof
	   KEY_DOWN		 of		       tipof
	   KEY_DOWN keypad-ed +	 of  Editor_Down       endof
	   KEY_UP		 of		       tipof
	   KEY_UP keypad-ed +	 of  Editor_Up	       endof
	   KEY_LEFT		 of		       tipof
	   KEY_LEFT keypad-ed +	 of  Editor_Left       endof
	   KEY_RIGHT		 of		       tipof
	   KEY_RIGHT keypad-ed + of  Editor_Right      endof
	   KEY_F10		 of  lStop? true!      endof
	   KEY_HELP		 of  Help	       endof
       endcase
       lStop? true?
   until
   glEditor? false!					       // Is Editor active ?
   graphics CellOffColor clear screen			       // Erase graphics screen (set with off cells)
   InfosLineNo clrline					       // Put right color for this line
   TRUE ShowGeneration					       // Show all cells, in graphics mode
   InfosLineNo clrline					       // Clear this line
   Frame						       // Redraw the 'frame' of the game
   ShowGenPop						       // Print the generation # and the nb of populations.
   ;  // CellsEditor


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : HeartLife												        \\
// 	 Stack : ( --- )												        \\
// Description : The 'heart' of the game.										        \\
//    Callers : life													        \\
// 	 Calls : KEY_HELP KEY_NEXT_GEN KEY_AUTO KEY_EDIT KEY_BYE NextGeneration AutomaticMode Help CellsEditor:2	        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  HeartLife
   CellsEditor						       // In first, we edit cells
   FALSE ilogical lStop?				       // Flag: stop main loop ?
   begin
       key						       // ( --- c )
       case
	   KEY_HELP	 of   Help	      endof
	   KEY_NEXT_GEN	 of   NextGeneration  endof
	   KEY_AUTO	 of   AutomaticMode   endof
	   KEY_EDIT	 of   CellsEditor     endof
	   KEY_BYE	 of   lStop? true!    endof
       endcase
       lStop? true?
   until
   ;  // HeartLife


// ---------------------------------------------------------------------------------------------------------------------------- \\
// 	  Word : life													        \\
// 	 Stack : ( --- )												        \\
// Description : The top word of the application									        \\
//    Callers :													        \\
// 	 Calls : C_NormalText cells changes InitThings Frame ZeroLife HeartLife						        \\
// ---------------------------------------------------------------------------------------------------------------------------- \\
:  life
   Screen C_NormalText ScreenBackground !		       // The screenbackground is Black
   GetVideoMode const nPrevVideoMode			       // Save previous video mode
   ?UpdateCursor true!					       // Update cursor when drawing
   graphics 640x480x16					       // Select graphics mode 640x480x16
   0 clear						       // Clear graphics screen
   InitThings						       // Initialize the game
   Frame						       // Print the frame
   ZeroLife						       // Empty all the cells.
   HeartLife						       // The heart of the game
   free cells						       // Free huge cells array
   free changes						       // Free another huge array
   screen GetVideoMode nPrevVideoMode != if		       // Restore previous video mode, if needed
       nPrevVideoMode SetVideoMode
   endif
   ;  // life


( Print code/data size )
( -------------------- )

infos


( Uncomment next section to build stand-alone application for MS/DOS (.exe) )
( ------------------------------------------------------------------------- )

" \\froth\\examples\\life.exe" save.exe
cr " LIFE.EXE app size is " type . "  bytes" type

